function [trn,sfx] = AO(s,thresh)
% [trn,sfx] = AO(s,a)
% Audio Oracle extends the algorithm of Factor Oracle for metric spaces over sequence of vectors s.
% It calls a dist function (scroll down) that currently implments an
% Euclidin norm.
%
% input:
% s - sequence of vectors. time orientation is along columns
% thresh - minimal allowed distance between symbols (default 0.1)
% output:
% trn - transition matrix (forward)
% sfx - suffix vector (backwards)

%Documentation:
% This version of FO uses a distance function (dts) do assign distances
% between an incoming symbol (vector) and symbols in the sequence so far.
% This comparison differs from the original FO where an incoming symbol was
% used to follow forward links (these are the links that have a symbol
% assigned to its arc, i.e. these are labeled transitions). Here the new
% measurement (symbol) could be compared in principle to all existing
% elements (symbols / vectors) in the sequnece, but we expoit here the
% FO construction algorithm to test the forward matches of the states that
% are visited during the suffix link passage. If the state reached by
% suffix link already has a forward transition that points to a new place
% in the sequence that is "similar enough" to the current symbol (i.e. that
% the distance between the states pointed to by the forward link from a state
% that was reached after following a suffix link), then a new forward
% transition is not added. Only in case that there are no similar symbols
% in the sequence (i.e. all symbols pointed to by the forward link are
% different from the new symbold by more then a certain predefined
% threhsold), then a new forward link is added to the new symbold, which is
% at the end of the sequence, i.e. the last appended symbol.
% The FO architecture allows to save many computation in comparison to
% complete comparison that would be done by methods such as DTW. Also, it
% incorporates the temporal strucutre by considering repeating suffixes
% using the suffix link structure. It should be noted that if several
% existing transitions were found to existing symbols in the sequence, then
% the algorithm will choose the best matching symbol, and create a
% suffix link to that location ONLY. This means that other locations in the
% sequence that have close but less similar next symbol will not be
% considered, which means that other potential suffix links are ignored.
% This of course raises the question of global optimality, since part of
% the ignored suffixes might have an overal (global) better match then the
% suffix chosen based on the last symbol comparison. One option to deal
% with the problem is to define a rank of similar states that would be
% retained during the search (the best two, three and etc.). This will
% expand the number of suffix links from every state by a constant factor
% equal to the rank (sfx(i) will become sfx(i,rank)), but the FO
% construction procedure will grow exponentially. (If rank is 2, then 
% the algorithm starts following two suffix links from the last state, 
% followed by four suffix links in the next step and so on). Hopefully most
% of the expanding branches will be termined since their symbols will exceed
% the distance treshold. 
% Question: what happens if we allow full rank suffixes? 
% Will it resemble approximate string matching with distance matrix?
% What is the tradeoff between global optimality based on local matching,
% which assumes pairwise similarities only and thus a markov model of order
% 2, versus the FO matching that takes into accound suffixes, but does not
% estimate global errors?

if nargin < 2,
thresh = 0.1;
end

trn = cell(size(s,2)+1,1);
sfx = zeros(1,size(s,2));

progressbar(0);

for i = 2:length(s)+1,
    progressbar(i/length(s));
    symbol = s(:,i-1); %symbol is a vector
    trn{i-1} = i; %forward step
    k = sfx(i-1);
    while k ~= 0,
        dvec = dts(trn{k},symbol,s);
        I = find(dvec<thresh);
        if isempty(I), %no transition with symbol from state k 
            trn{k} = [trn{k} i]; %append (add) i to set of transitions from state k
            k = sfx(k); %follow suffix link
        else
            break
        end
    end
    if k == 0, %if no transition exists
        sfx(i) = 1; %point to state 1 (beginning)
    else 
        % choose first transition from k that is < thresh as suffix
        %sfx(i) = trn{k}(min(I)); 
        % another option: choose suffix link to state with minimal distance
        [val,ival] = min(dvec(I));
        sfx(i) = trn{k}(I(ival));
    end
end
trn{end} = [2:length(s)]; %loop? last state point to any state?

function dist = dts(fwd,symbol,s)
% vector of distances between new symbol and all states 
% pointed to by fwd transition links.
% The distance is normalized by the dimension of the feature vector

D = length(symbol); %dimension of the feature vector

for i = 1:length(fwd),
    dist(i) = norm(s(:,fwd(i)-1)-symbol); %fwd(i)-1 since fwd indexes are shifted by one (start with 1 and not zero like in the original FO).
end

dist = dist/D;